#include "sdbf_class.h"
#include "sdbf_defines.h"
#include "sdhash-srv.h"
#include "set_list.h"

#include <unordered_set>
#include <boost/regex.hpp>
#include <iostream>
#include <fstream>
#include <string>

#define CHUNK_SIZE 128*MB
 
// Global parameters
//extern sdbf_parameters_t sdbf_sys;

// State
static pthread_t *thread_pooll = NULL; // file-local thread pool

// State
static sdbf_set **set_list= NULL;
static char **name_list= NULL;
static uint32_t curr_sdbf = 0;
static pthread_mutex_t set_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t hash_mutex = PTHREAD_MUTEX_INITIALIZER;


/**
 * Initialization of SDBF structures. Must be called once before the remaining sdbf functions are used.
 */
int set_list_init() {
	set_list = (sdbf_set **)alloc_check( ALLOC_ZERO, (MAX_FILES*sizeof( sdbf_set **)), "set_list_init", "set_list", ERROR_EXIT);
	name_list = (char **)alloc_check( ALLOC_ZERO, (MAX_FILES*sizeof( char**)), "set_list_init", "name_list", ERROR_EXIT);
	return 0;
}


/**
 * Frees up SDBF structures. 
 */
void set_list_free() {
	if( set_list)
			free( set_list);
	if( name_list)
			free( name_list);
}

/**
* add set and its name to the set list.
*/
int add_set( sdbf_set *set, char *name) {
	assert( set_list && curr_sdbf < MAX_FILES);

	pthread_mutex_lock( &set_mutex);    
		set_list[curr_sdbf] = set;
		name_list[curr_sdbf] = name;
		curr_sdbf++;
	pthread_mutex_unlock( &set_mutex);  
	return curr_sdbf;
}

/**
 * Returns the number of sets in our list
 */
int get_set_count() {
	return curr_sdbf;
}

/**
 * Returns the set associated with an index
 */
sdbf_set *get_set( uint32_t index) {
	if( index < curr_sdbf)
		return set_list[index];
	else
		return NULL;
}

/**
 * Returns the name (string) associates with a set
 */
char *get_set_name( uint32_t index) {
	if( index < curr_sdbf)
		return name_list[index];
	else
		return NULL;
}

// NOT in sdbf class
/** 
	thread "process" function for file-by-file hashing
*/

void
*thread_sdbf_hashfile( void *task_param) {
    filehash_task_t *task = (filehash_task_t *)task_param;
    struct stat file_stat;
    int32_t chunks = 0,csize = 0;
    ifstream *is = new ifstream();

    int i,j;
    for( i=task->tid; i<task->file_count; i+=task->tcount) {
        if (stat(task->filenames[i],&file_stat))
            continue;
        is->open(task->filenames[i], ios::binary);
		try {
			class sdbf *sdbfm = new sdbf(task->filenames[i],is,0,file_stat.st_size);
			task->addset->add(sdbfm);
		} catch (int e) {
			if (e==-2)
			   exit(-2);
		}
        is->close();
    }
}



/**
 * Compute SD for a list of files & add them to a new set.
 */

sdbf_set
*sdbf_hash_files( char **filenames, uint32_t file_count ) {
    int32_t i, t, j, thread_cnt = 1;
    sdbf_set *myset = new sdbf_set();
    struct stat file_stat;
    int32_t chunks = 0,csize = 0;
    ifstream *is = new ifstream();

    // Sequential implementation
    if( thread_cnt == 1) {
        for( i=0; i<file_count; i++) {
            if (stat(filenames[i],&file_stat))
                continue;
            is->open(filenames[i], ios::binary);
			try {
				class sdbf *sdbfm = new sdbf(filenames[i],is,0,file_stat.st_size);
				myset->add(sdbfm);
			} catch (int e) {
				if (e==-2)
				   exit(-2);
			}
            is->close();
        }
    // Threaded implementation
    } else {
        pthread_t *thread_pooll = (pthread_t *) alloc_check( ALLOC_ZERO, thread_cnt*sizeof( pthread_t), "sdbf_hash_files", "thread_pool", ERROR_EXIT);
        filehash_task_t *tasks = (filehash_task_t *) alloc_check( ALLOC_ZERO, thread_cnt*sizeof( filehash_task_t), "sdbf_hash_files", "tasks", ERROR_EXIT);
        for( t=0; t<thread_cnt; t++) {
            tasks[t].tid = t;
            tasks[t].tcount = thread_cnt;
            tasks[t].filenames = filenames;
            tasks[t].file_count = file_count;
            tasks[t].addset = myset;
            if( pthread_create( &thread_pooll[t], NULL, thread_sdbf_hashfile, (void *)(tasks+t) )) {
                fprintf( stderr, "ERROR: Could not create thread.\n");
                exit(-1);
            }
        }
        for( t=0; t<thread_cnt; t++) {
            pthread_join( thread_pooll[t], NULL);
        }
    // End threading
    }
    return myset;
}

/**
 * Compute SD for a filename containing a list of files & add them to a new set.
 */
sdbf_set*
hash_list(char *listfn, uint32_t dd_block_size) {
	sdbf_set *set1 = new sdbf_set();
	int i;
	struct stat stat_res;
	if( stat(listfn, &stat_res) != 0) {
		fprintf( stderr, "ERROR: Could not access input file \"%s\". Exiting.\n",listfn);
		return set1; 
	}
	char *filename;
	char **filelist;
	filelist=(char**)alloc_check(ALLOC_ONLY,stat_res.st_size,"main", "filenames file block",ERROR_EXIT);
	processed_file_t *mlist=process_file(listfn, 1, 0);
	if (!mlist) {
		fprintf( stderr, "ERROR: Could not open input file \"%s\". Exiting.\n",listfn); 
		return set1;
	}
	i=0;
	while (filename = strsep( (char**)&mlist->buffer, "\n") ) {
		filelist[i]=filename;
		i++;
	}

	pthread_mutex_unlock( &hash_mutex);    
	if (dd_block_size > 0)  {
		set1->add_and_hash(filelist, i-1, dd_block_size*KB, 128*MB);
	} else {
		set1=sdbf_hash_files( filelist, i-1);
	}
	pthread_mutex_unlock( &hash_mutex);    
	return set1;
}
